(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) :
  typeof define === 'function' && define.amd ? define(['exports'], factory) :
  (factory((global.BasicStreams = global.BasicStreams || {})));
}(this, function (exports) { 'use strict';

  function __commonjs(fn, module) { return module = { exports: {} }, fn(module, module.exports), module.exports; }

  var _slicedToArray = function () { function sliceIterator(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"]) _i["return"](); } finally { if (_d) throw _e; } } return _arr; } return function (arr, i) { if (Array.isArray(arr)) { return arr; } else if (Symbol.iterator in Object(arr)) { return sliceIterator(arr, i); } else { throw new TypeError("Invalid attempt to destructure non-iterable instance"); } }; }();

  // We're not sure yet if this function should be public
  function fromArray(xs) {
    return function (sink) {
      xs.forEach(function (x) {
        sink(x);
      });
      return function () {};
    };
  }

  /* Given a loose basic-stream, that obeys at least following laws:
   *   1. Stream is a function,
   *   2. It accepts one argument, the subscriber function (aka `sink`);
   * returns a valid basic-steam, that obeys all laws of a stream from protocol,
   * and also immune to violations of any of usage laws except:
   *   2. `sink` must be a function
   */

  // A helper to utilize Flow's Disjoint Unions
  // http://flowtype.org/blog/2015/07/03/Disjoint-Unions.html
  //
  function fromLoose(looseStream) {
    return function (_sink) {
      var sink = _sink;
      var disposer = looseStream(function (x) {
        // fix usage #1
        if (sink !== null) {
          // fix stream #6
          sink(x); // fix stream #4
        }
        // fix usage #3
      });
      return function () {
        // fix stream #3
        if (sink === null) {
          // fix usage #5
          return;
        }
        sink = null;
        if (typeof disposer === 'function') {
          // fix stream #3
          disposer(); // fix usage #4
        }
        disposer = null;
        // fix stream #5
      };
    };
  }

  /* Represents an empty stream
   */
  var empty = function () {
    return function () {};
  };

  /* Creates a stream containing given value
   */
  function just(x) {
    return function (sink) {
      sink(x);
      return function () {};
    };
  }

  /* Lifts function `A => B` to a function that operates
   * on streams `Stream<A> => Stream<B>`
   */
  function map(fn) {
    return function (stream) {
      return function (sink) {
        return stream(function (payload) {
          return sink(fn(payload));
        });
      };
    };
  }

  /* Given a predicate `A => boolean` returns a function
   * that operates on streams `Stream<A> => Stream<A>`.
   * The result function returns a stream without values that don't satisfy predicate.
   */
  function filter(predicate) {
    return function (stream) {
      return function (sink) {
        return stream(function (payload) {
          if (predicate(payload)) {
            sink(payload);
          }
        });
      };
    };
  }

  /* Given a function `A => Stream<B>` returns a function
   * that operates on streams `Stream<A> => Stream<B>`.
   * The result function will spawn a `Stream<B>`
   * for each value from `Stream<A>` using the provided function.
   * The final `Stream<B>` will contain values from all spawned streams.
   */
  function chain(fn) {
    return function (stream) {
      return function (sink) {
        var spawnedDisposers = [];
        var mainDisposer = stream(function (payload) {
          spawnedDisposers.push(fn(payload)(sink));
        });
        return function () {
          spawnedDisposers.forEach(function (fn) {
            return fn();
          });
          mainDisposer();
        };
      };
    };
  }

  /* Same as `chain()`, except the final stream will contain
   * only that values from each spawned streams that was
   * emitted before the next stream was spawned.
   */
  function chainLatest(fn) {
    return function (stream) {
      return function (sink) {
        var spawnedDisposers = function () {};
        var mainDisposer = stream(function (payload) {
          spawnedDisposers();
          spawnedDisposers = fn(payload)(sink);
        });
        return function () {
          spawnedDisposers();
          mainDisposer();
        };
      };
    };
  }

  /* Given a stream of functions `Stream<A => B>`, returns a function
   * that operates on streams `Stream<A> => Stream<B>`.
   * The result stream `Stream<B>` will contain values created by applying
   * the latest function from `Stream<A => B>` to the latest value from `Stream<A>`
   * every time one of them updates.
   */
  function ap(streamf) {
    return function (streamv) {
      return function (sink) {
        var latestF = { type: 'nothing' };
        var latestV = { type: 'nothing' };
        var push = function () {
          if (latestF.type === 'just' && latestV.type === 'just') {
            var fn = latestF.value;
            sink(fn(latestV.value));
          }
        };
        var disposef = streamf(function (f) {
          latestF = { type: 'just', value: f };
          push();
        });
        var disposev = streamv(function (v) {
          latestV = { type: 'just', value: v };
          push();
        });
        return function () {
          disposef();
          disposev();
        };
      };
    };
  }

  /* Lifts a 2 arity function `(A, B) => C` to a function that operates
   * on streams `(Stream<A>, Stream<B>) => Stream<C>`
   */
  function map2(fn) {
    return function (sA, sB) {
      var ofFn = map(function (a) {
        return function (b) {
          return fn(a, b);
        };
      })(sA);
      return ap(ofFn)(sB);
    };
  }

  /* Lifts a 3 arity function `(A, B, C) => D` to a function that operates
   * on streams `(Stream<A>, Stream<B>, Stream<C>) => Stream<D>`
   */
  function map3(fn) {
    return function (sA, sB, sC) {
      var ofFn = map(function (a) {
        return function (b) {
          return function (c) {
            return fn(a, b, c);
          };
        };
      })(sA);
      return ap(ap(ofFn)(sB))(sC);
    };
  }

  /* Merges several streams of same type to a single stream of that type.
   * The result stream will contain values from all streams
   */
  function merge(streams) {
    return chain(function (x) {
      return x;
    })(fromArray(streams));
  }

  /* Given a reducer function `(B, A) => B`, and a seed value of type B,
   * returns a function that operates on streams `Stream<A> => Stream<B>`.
   * Will apply reducer to each value in `Stream<A>` and push the result to `Stream<B>`
   */
  function scan(reducer, seed) {
    return function (stream) {
      return function (sink) {
        var current = seed;
        sink(current);
        return stream(function (x) {
          current = reducer(current, x);
          sink(current);
        });
      };
    };
  }

  /* Given a number N, returns a function that operates on streams `Stream<A> => Stream<A>`.
   * The result stream will contain only first N items from source stream
   */
  function take(n) {
    return function (stream) {
      return function (sink) {
        var count = 0;
        var disposer = null;
        var dispose = function () {
          if (disposer !== null) {
            disposer();
            disposer = null;
          }
        };
        disposer = stream(function (x) {
          count++;
          if (count <= n) {
            sink(x);
          }
          if (count >= n) {
            dispose();
          }
        });
        if (count >= n) {
          dispose();
        }
        return dispose;
      };
    };
  }

  /* Given a predicate `A => boolean` returns a function
   * that operates on streams `Stream<A> => Stream<A>`.
   * The result stream will contain values from source stream
   * before the first value that doesn't satisfy predicate.
   */
  function takeWhile(pred) {
    return function (stream) {
      return function (sink) {
        var completed = false;
        var disposer = null;
        var dispose = function () {
          if (disposer !== null) {
            disposer();
            disposer = null;
          }
        };
        disposer = stream(function (x) {
          if (!completed) {
            if (pred(x)) {
              sink(x);
            } else {
              completed = true;
              dispose();
            }
          }
        });
        if (completed) {
          dispose();
        }
        return dispose;
      };
    };
  }

  /* Given a controller stream, returns a function that operates
   * on streams `Stream<A> => Stream<A>`. The result stream will contain values
   * from source stream until the first value from controller stream.
   */
  function takeUntil(controller) {
    return function (stream) {
      return function (sink) {
        var mainDisposer = null;
        var ctrlDisposer = null;
        var dispose = function () {
          if (mainDisposer !== null) {
            mainDisposer();
            mainDisposer = null;
          }
          if (ctrlDisposer !== null) {
            ctrlDisposer();
            ctrlDisposer = null;
          }
        };

        mainDisposer = stream(sink);
        ctrlDisposer = controller(dispose);

        // in case controller cb was called sync before we obtained ctrlDisposer
        if (mainDisposer === null) {
          dispose();
        }

        return dispose;
      };
    };
  }

  /* Given a number N, returns a function that operates on streams `Stream<A> => Stream<A>`.
   * The result stream will contain only items from source starting from (N+1)th one
   */
  function skip(n) {
    return function (stream) {
      return function (sink) {
        var count = 0;
        return stream(function (x) {
          count++;
          if (count > n) {
            sink(x);
          }
        });
      };
    };
  }

  /* Given a predicate `A => boolean` returns a function
   * that operates on streams `Stream<A> => Stream<A>`.
   * The result stream will contain values from source stream
   * starting from the first that doesn't satisfy predicate.
   */
  function skipWhile(pred) {
    return function (stream) {
      return function (sink) {
        var started = false;
        return stream(function (x) {
          if (!started) {
            started = !pred(x);
          }
          if (started) {
            sink(x);
          }
        });
      };
    };
  }

  /* Given a comparator function `(A, A) => boolean` returns a function that
   * operates on streams `Stream<A> => Stream<A>`. For each element X from
   * the source stream (except the first one) calls comparator with (Y, X)
   * as arguments, where Y is the last element in the result stream.
   * If comparator returns false, X goes to the result stream.
   */
  function skipDuplicates(comp) {
    return function (stream) {
      return function (sink) {
        var latest = { type: 'nothing' };
        return stream(function (x) {
          if (latest.type === 'nothing' || !comp(latest.value, x)) {
            latest = { type: 'just', value: x };
            sink(x);
          }
        });
      };
    };
  }

  /* Given a stream returns a new stream of same type. The new stream will
   * have at most one subscription at any given time to the original stream.
   * It allows you to connect several subscribers to a stream using only one subscription.
   */
  function multicast(stream) {
    var sinks = [];
    var push = function (x) {
      sinks.forEach(function (sink) {
        if (sinks.indexOf(sink) !== -1) {
          sink(x);
        }
      });
    };
    var unsub = null;
    return function (sink) {
      var disposed = false;
      sinks = [].concat(sinks, [sink]);
      if (sinks.length === 1) {
        unsub = stream(push);
      }
      return function () {
        if (disposed) {
          return;
        }
        disposed = true;
        var index = sinks.indexOf(sink);
        sinks = [].concat(sinks.slice(0, index), sinks.slice(index + 1, sinks.length));
        if (sinks.length === 0 && unsub !== null) {
          unsub();
          unsub = null;
        }
      };
    };
  }

  /* Given a value of type `A` returns a function that
   * operates on streams `Stream<A> => Stream<A>`.
   * The result stream is a copy of source stream
   * with the given value added at the beginning.
   */
  function startWith(x) {
    return function (stream) {
      return function (sink) {
        sink(x);
        return stream(sink);
      };
    };
  }

  /* Given an array of streams returns a stream of arrays.
   * This is basically an implementation of FantasyLand's sequence() for Array
   * specialized to Streams.
   */
  function combineArray(arr) {
    return arr.reduce(map2(function (arr, i) {
      return arr.concat([i]);
    }), just([]));
  }

  var fromPairsLifted = map(function (pairs) {
    var result = {};
    pairs.forEach(function (_ref) {
      var _ref2 = _slicedToArray(_ref, 2);

      var key = _ref2[0];
      var value = _ref2[1];

      result[key] = value;
    });
    return result;
  });

  /* Given an object (a.k.a map/hash) of streams returns a stream of objects.
   * Same as combineArray but for objects.
   */
  function combineObject(obj) {
    var ofPairs = Object.keys(obj).map(function (key) {
      return map(function (x) {
        return [key, x];
      })(obj[key]);
    });
    return fromPairsLifted(combineArray(ofPairs));
  }

  /* Given a transducer that conforms [Protocol][1],
   * returns a function that operates on streams `Stream<A> => Stream<B>`.
   *
   * [1]: https://github.com/cognitect-labs/transducers-js#the-transducer-protocol
   */
  function transduce(transducer) {
    return function (stream) {
      return function (sink) {
        var thisDisposed = false;
        var sourceDisposer = null;

        var transformer = transducer({
          '@@transducer/result': function () {},
          '@@transducer/step': function (result, input) {
            if (!thisDisposed) {
              sink(input);
            }
            return result;
          }
        });

        var disposeSource = function () {
          if (sourceDisposer !== null) {
            sourceDisposer();
            sourceDisposer = null;
          }
        };

        sourceDisposer = stream(function (x) {
          if (transformer === null) {
            return;
          }

          // it returns either null or tReduced<null>
          if (null !== transformer['@@transducer/step'](null, x)) {
            transformer['@@transducer/result'](null);
            transformer = null;
            disposeSource();
          }
        });

        if (transformer === null) {
          disposeSource();
        }

        return function () {
          thisDisposed = true;
          disposeSource();
        };
      };
    };
  }

  var index = __commonjs(function (module) {
    module.exports = {
      equals: 'equals',
      concat: 'concat',
      empty: 'empty',
      map: 'map',
      ap: 'ap',
      of: 'of',
      reduce: 'reduce',
      sequence: 'sequence',
      chain: 'chain',
      extend: 'extend',
      extract: 'extract'
    };
  });

  var fl = index && typeof index === 'object' && 'default' in index ? index['default'] : index;

  function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

  var Stream = function () {

    // Example:
    //
    //    const myWrappedStream = new Stream(sink => {
    //      sink(1)
    //      retrun () => {}
    //    })
    //

    function Stream(basicStream) {
      _classCallCheck(this, Stream);

      // The wrapped BasicStream is available as `wrapped.observe`.
      // You can call it as a method in order to subscribe to the stream:
      //
      //    const usub = wrapped.observe(callback)
      //    unsub()
      //
      // Or you can use `wrapped.observe` to extract BasicStream from the wrapper e.g.:
      //
      //    const basicStream2 = map(x => x * 2)(wrapped.observe)
      //
      this.observe = fromLoose(basicStream);
    }

    // Semigroup

    Stream.prototype.concat = function concat(other) {
      return new Stream(merge([this.observe, other.observe]));
    };

    // Monoid

    Stream.empty = function empty$$() {
      return new Stream(empty);
    };

    Stream.prototype.empty = function empty$$() {
      return Stream.empty();
    };

    // Functor

    Stream.prototype.map = function map$$(f) {
      return new Stream(map(f)(this.observe));
    };

    // Apply

    Stream.prototype.ap = function ap$$(other) {
      // In order for .ap() to work `this` must be a `Stream<A => B>`,

      var streamF = this.observe;
      return new Stream(ap(streamF)(other.observe));
    };

    // Applicative

    Stream.of = function of(x) {
      return new Stream(just(x));
    };

    Stream.prototype.of = function of(x) {
      return Stream.of(x);
    };

    // Chain

    Stream.prototype.chain = function chain$$(f) {
      return new Stream(chain(function (x) {
        return f(x).observe;
      })(this.observe));
    };

    // ------------------ Not from FL ------------------

    Stream.prototype.filter = function filter$$(f) {
      return new Stream(filter(f)(this.observe));
    };

    Stream.prototype.scan = function scan$$(f, seed) {
      return new Stream(scan(f, seed)(this.observe));
    };

    Stream.prototype.chainLatest = function chainLatest$$(f) {
      return new Stream(chainLatest(function (x) {
        return f(x).observe;
      })(this.observe));
    };

    Stream.prototype.take = function take$$(n) {
      return new Stream(take(n)(this.observe));
    };

    Stream.prototype.takeWhile = function takeWhile$$(f) {
      return new Stream(takeWhile(f)(this.observe));
    };

    Stream.prototype.takeUntil = function takeUntil$$(other) {
      return new Stream(takeUntil(other.observe)(this.observe));
    };

    Stream.prototype.skip = function skip$$(n) {
      return new Stream(skip(n)(this.observe));
    };

    Stream.prototype.skipWhile = function skipWhile$$(f) {
      return new Stream(skipWhile(f)(this.observe));
    };

    Stream.prototype.skipDuplicates = function skipDuplicates$$(f) {
      return new Stream(skipDuplicates(f)(this.observe));
    };

    Stream.prototype.multicast = function multicast$$() {
      return new Stream(multicast(this.observe));
    };

    Stream.prototype.startWith = function startWith$$(x) {
      return new Stream(startWith(x)(this.observe));
    };

    Stream.prototype.transduce = function transduce$$(transducer) {
      return new Stream(transduce(transducer)(this.observe));
    };

    Stream.merge = function merge$$(streams) {
      return new Stream(merge(streams.map(function (x) {
        return x.observe;
      })));
    };

    Stream.map2 = function map2$$(f) {
      var lifted = map2(f);
      return function (sA, sB) {
        return new Stream(lifted(sA.observe, sB.observe));
      };
    };

    Stream.map3 = function map3$$(f) {
      var lifted = map3(f);
      return function (sA, sB, sC) {
        return new Stream(lifted(sA.observe, sB.observe, sC.observe));
      };
    };

    Stream.combineArray = function combineArray$$(arr) {
      return new Stream(combineArray(arr.map(function (x) {
        return x.observe;
      })));
    };

    Stream.combineObject = function combineObject$$(obj) {
      var ofBasicStreams = {};
      Object.keys(obj).forEach(function (key) {
        ofBasicStreams[key] = obj[key].observe;
      });
      return new Stream(combineObject(ofBasicStreams));
    };

    return Stream;
  }();

  // Flow doesn't support computed method names,
  // we have to bypass type checking for FL methods :(
  //
  function addFlMethods(constructor, proto) {
    ['concat', 'empty', 'map', 'ap', 'of', 'chain'].forEach(function (name) {
      proto[fl[name]] = proto[name];
    });['empty', 'of'].forEach(function (name) {
      constructor[fl[name]] = constructor[name];
    });
  }
  addFlMethods(Stream, Stream.prototype);

var fantasy = Object.freeze({
    Stream: Stream
  });

  /* Pipes a value through a line of functions.
   * Allows to replace `f3(f2(f1(x)))` with `pipe(x, f1, f2, f3)`
   *
   * Note: It's not well typed with Flow, we had to resort to `any`
   */
  function pipe(value) {
    for (var _len = arguments.length, fns = Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
      fns[_key - 1] = arguments[_key];
    }

    return fns.reduce(function (x, fn) {
      return fn(x);
    }, value);
  }

var utils = Object.freeze({
    pipe: pipe
  });

  exports.fantasy = fantasy;
  exports.utils = utils;
  exports.fromLoose = fromLoose;
  exports.empty = empty;
  exports.just = just;
  exports.map = map;
  exports.filter = filter;
  exports.chain = chain;
  exports.chainLatest = chainLatest;
  exports.ap = ap;
  exports.map2 = map2;
  exports.map3 = map3;
  exports.merge = merge;
  exports.scan = scan;
  exports.take = take;
  exports.takeWhile = takeWhile;
  exports.takeUntil = takeUntil;
  exports.skip = skip;
  exports.skipWhile = skipWhile;
  exports.skipDuplicates = skipDuplicates;
  exports.multicast = multicast;
  exports.startWith = startWith;
  exports.combineArray = combineArray;
  exports.combineObject = combineObject;
  exports.transduce = transduce;

}));